package pipe.dataLayer;

// Collections
import java.awt.Color;
import java.io.File;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Observable;
import java.util.Random;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import pipe.gui.CreateGui;

/**
 * <b>DataLayer</b> - Encapsulates entire Petri-Net, also contains
 * functions to perform calculations
 *
 * @see <p><a href="..\PNMLSchema\index.html">PNML  -
 * Petri-Net XMLSchema (stNet.xsd)</a>
 * @see </p><p><a href="uml\DataLayer.png">DataLayer UML</a></p>
 * @version 1.0
 * @author James D Bloom
 * @author David Patterson Jan 2, 2006: Changed the fireRandomTransition 
 * method to give precedence to immediate transitions.
 * 
 * @author Edwin Chung added a boolean attribute to each matrix generated to
 * prevent them from being created again when they have not been changed (6th Feb 2007)
  * 
 * @author Ben Kirby Feb 10, 2007: Removed savePNML method and the createPlaceElement,
 * createAnnotationElement, createArcElement, createArcPoint, createTransitionElement
 * methods it uses to a separate DataLayerWriter class, as part of refactoring to 
 * remove XML related actions from the DataLayer class.
 * 
 * @author Ben Kirby Feb 10, 2007: Split loadPNML into two bits. All XML work (Files, 
 * transformers, documents) is done in new PNMLTransformer class. The calls to actually 
 * populate a DataLayer object with the info contained in the PNML document have been 
 * moved to a createFromPNML method. The DataLayer constructor which previously used 
 * loadPNML has been changed to reflect these modifications. Also moved getDOM methods 
 * to PNMLTranformer class, as getDom is XML related. Removed getDom() (no arguments) 
 * completely as this is not called anywhere in the application. 
 *
 *@author Will Master Feb 13 2007: Added methods getPlacesCount and getTransitionsCount
 * to avoid needlessly copying place and transition arrayLists.
 * 
 * @author Edwin Chung 15th Mar 2007: modified the createFromPNML function so that
 * DataLayer objects can be created outside GUI
 * 
 * @author Dave Patterson 24 April 2007: Modified the fireRandomTransition method so it is
 * quicker when there is only one transition to fire (just fire it, don't get a random
 * variable first). Also, throw a RuntimeException if a rate less than 1 is detected.
 * The current code uses the rate as a weight, and a rate such as 0.5 leads to a 
 * condition like that of bug 1699546 where no transition is available to fire.
 * 
 * @author Dave Patterson 10 May 2007: Modified the fireRandomTransitino method
 * so it now properly handles fractional weights. There is no RuntimeException
 * thrown now. The code for timed transitions uses the same logic, but will soon
 * be changed to use exponentially distributed times where fractional rates 
 * are valid.
 * 
 **/
public class DataLayer extends Observable implements pipe.gui.Constants{

  private static Random randomNumber = new Random(); // Random number generator

  /** PNML File Name */
  public static String pnmlName = null;

  /** List containing all the Place objects in the Petri-Net */
  private ArrayList placesArray = null;
  /** ArrayList containing all the Transition objects in the Petri-Net */
  private ArrayList transitionsArray = null;
  /** ArrayList containing all the Arc objects in the Petri-Net */
  private ArrayList arcsArray = null;
  /** ArrayList containing all the Token objects in the Petri-Net */
  private ArrayList tokensArray = null;
  /** ArrayList containing all the Arrow objects in the Petri-Net */
  private ArrayList arrowsArray = null;

  /** ArrayList for net-level label objects (as opposed to element-level labels).*/
  private ArrayList labelsArray = null;

  /** An ArrayList used to point to either the Arc, Place or Transition ArrayLists when these ArrayLists are being update */
  private ArrayList changeArrayList = null;

  /** Initial Markup Matrix */
  private int[] initialMarkupMatrix = null;
  /** Initial Markup Matrix */
  private int[] currentMarkupMatrix = null;
  /** Markup Matrix Storage used during animation */
  private int[] markupMatrixAnimationStorage = null;

  /** Markup Matrix Storage */
  private Object[] markupStore = {null, null, null, null, null, null, null, null, null, null};
  /** Markup Matrix Storage Position */
  private int position = 0;

  /** Markup Storage Array */
  private PNMatrix markupHistoryMatrix = null;
  /** Foward Incidence Matrix */
  private PNMatrix fowardsIncidenceMatrix = null;
  /** Backward Incidence Matrix */
  private PNMatrix backwardsIncidenceMatrix = null;
  /** Incidence Matrix */
  private PNMatrix incidenceMatrix = null;
 // /** Used to determine whether the matrixes have been modified */
  static boolean initialMarkupMatrixChanged = true;
  
  static boolean currentMarkupMatrixChanged = true;

  /** X-Axis Scale Value */
  private final int DISPLAY_SCALE_FACTORX = 7; // Scale factors for loading other Petri-Nets (not yet implemented)
  /** Y-Axis Scale Value */
  private final int DISPLAY_SCALE_FACTORY = 7; // Scale factors for loading other Petri-Nets (not yet implemented)
  /** X-Axis Shift Value */
  private final int DISPLAY_SHIFT_FACTORX = 270; // Scale factors for loading other Petri-Nets (not yet implemented)
  /** Y-Axis Shift Value */
  private final int DISPLAY_SHIFT_FACTORY = 120; // Scale factors for loading other Petri-Nets (not yet implemented)

  /** Hashtable which maps PlaceTransitionObjects to their list of connected arcs */
  private Hashtable arcsMap = null;

  /**
   * Create Petri-Net object from PNML file with URI pnmlFileName
   *
   *
   * @param pnmlFileName Name of PNML File
   */
  public DataLayer(String pnmlFileName) {
    initializeMatrixes();
    PNMLTransformer transform = new PNMLTransformer();
    File temp = new File(pnmlFileName);
    pnmlName = temp.getName();
    createFromPNML(transform.transformPNML(pnmlFileName));

	
	

	//BK: no need for these catches now - these - all caught in method calls now

//	} catch (IOException e) {
//	System.err.println("IOException thrown in dataLayer(String pnmlFileName) : dataLayer Class : dataLayer Package");
//	e.printStackTrace(System.err);
//	} catch (SAXException e) {
//	System.err.println("SAXException thrown in dataLayer(String pnmlFileName) : dataLayer Class : dataLayer Package");
//	e.printStackTrace(System.err);
//	} catch (TransformerException e) {
//	System.err.println("TransformerException thrown in dataLayer(String pnmlFileName) : dataLayer Class : dataLayer Package");
//	e.printStackTrace(System.err);
//	} catch (ParserConfigurationException e) {
//	System.err.println("ParserConfigurationException thrown in dataLayer(String pnmlFileName) : dataLayer Class : dataLayer Package");
//	e.printStackTrace(System.err);
//	}
}

  /**
   * Create Petri-Net object from pnmlFile
   *
   * @param pnmlFile PNML File
   */
  public DataLayer(File pnmlFile) {
      this(pnmlFile.getAbsolutePath());
  }

  /**
   * Create empty Petri-Net object
   */
  public DataLayer() {
    initializeMatrixes();
  }

  /**
   * Initialize Arrays
   */
  private void initializeMatrixes() {
    placesArray = new ArrayList();
    transitionsArray = new ArrayList();
    arcsArray = new ArrayList();
    tokensArray = new ArrayList();
    arrowsArray = new ArrayList();
    labelsArray = new ArrayList();
    initialMarkupMatrix = null;
    fowardsIncidenceMatrix = null;
    backwardsIncidenceMatrix = null;
    incidenceMatrix = null;

    // may as well do the hashtable here as well
    arcsMap = new Hashtable();
  }

  /* TK - this method is never called.
   * Populates the arcsMap hashtable enabling easier cross referencing of places, transitions and the arcs connected to them.
   *
   
  private void setArcConnectionMap() {
    // map (PTO, arcslist)


    // get all the PlaceTransitionObjects
    ArrayList allPTO = new ArrayList(placesArray);
    allPTO.addAll(transitionsArray);

    Iterator pto;
    Iterator arcs;
    PlaceTransitionObject current ;

    pto = allPTO.iterator();
    while (pto.hasNext()) {
      current = (PlaceTransitionObject)pto.next();
      // make an entry for each pto in our mapping structure
      arcsMap.put(current,new ArrayList());
    }

    arcs = arcsArray.iterator();
    Arc currentArc;
    PlaceTransitionObject source;
    PlaceTransitionObject target;
    ArrayList arcslist = null;


    // iterate over the arcs, getting the source and destination ptos and adding the arc to their lists
    while (arcs.hasNext()) {
      currentArc = (Arc)arcs.next();
      source = currentArc.getSource();
      target = currentArc.getTarget();


      // get the list of arcs attached to the source and target and add the arc to them
      try {
        ((ArrayList)arcsMap.get(source)).add(currentArc);
      } catch (NullPointerException ne1) {
//        System.out.println("Populating arcsMap: " + ne1.toString());
      }

      try {
        ((ArrayList)arcsMap.get(target)).add(currentArc);
      } catch (NullPointerException ne2) {
//        System.out.println("Populating arcsMap: " + ne2.toString());
      }



    }

  }
  */


 /* TK - this method is never called
  * Remove first Place that has an id equal to idInput
  *
  * @param idInput id of Place object to remove from Petri-Net
  *
  public void removePlace(String idInput) {
    for(int i = 0 ; i < placesArray.size(); i++)
      if(idInput.equals(((Place)placesArray.get(i)).getId())) {
    placesArray.remove(i);
    setChanged();
    setMatrixChanged();
      }
  }
  */

  /* TK - this method is never called
   * Remove first Transition that has an id equal to idInput
   *
   * @param idInput id of Transition object to remove from Petri-Net
   *
  public void removeTransition(String idInput) {
    for(int i = 0 ; i < transitionsArray.size(); i++)
      if(idInput.equals(((Transition)transitionsArray.get(i)).getId())) {
    transitionsArray.remove(i);
    setChanged();
    setMatrixChanged();
      }
  }
  */

  /* TK - this method is never called
   * Remove first Arc that has an id equal to idInput
   *
   * @param idInput id of Arc object to remove from Petri-Net
   *
  public void removeArc(String idInput) {

    for(int i = 0 ; i < arcsArray.size(); i++)
      if(idInput.equals(((Arc)arcsArray.get(i)).getId())) {
    arcsArray.remove(i);
    setChanged();
    setMatrixChanged();
      }
  }
  */


  /**
   * Add placeInput to the back of the Place ArrayList
   * All observers are notified of this change (Model-View Architecture)
   *
   * @param placeInput Place Object to add
   */
  private void addPlace(Place placeInput) {
    boolean unique = true;
    if(placeInput != null) {
      if(placeInput.getId() != null && placeInput.getId().length() > 0) {
        for(int i = 0 ; i < placesArray.size() ; i++) {
          if(placeInput.getId().equals(((Place)placesArray.get(i)).getId())) {
            unique = false;
          }
        }
      } else {
        String id = null;
        if(placesArray != null && placesArray.size() > 0) {
          int no = placesArray.size();
          //          id = "P" + no;
          do {
            // System.out.println("in while loop");
            for(int i = 0 ; i < placesArray.size() ; i++) {
              id = "P" + no;
              if(placesArray.get(i) != null) {
                if(id.equals(((Place)placesArray.get(i)).getId())) {
                  // System.out.println("testing id: " + id);
                  unique = false;
                  no++;
                } else {
                  unique = true;
                }
              }
            }
          } while(!unique);
        } else {
          id = "P0";
        }

        if(id != null) {
          placeInput.setId(id);
        } else {
          placeInput.setId("error");
        }
      }
      placesArray.add(placeInput);
      setChanged();
      setMatrixChanged();
      // notifyObservers(placeInput.getBounds());
      notifyObservers(placeInput);
    }
  }

  /**
   * Add placeInput to the back of the Place ArrayList
   * All observers are notified of this change (Model-View Architecture)
   *
   * @param placeInput Place Object to add
   */
  private void addAnnotation(AnnotationNote labelInput) {
  	boolean unique = true;
  	labelsArray.add(labelInput);
  	setChanged();
  	notifyObservers(labelInput);
  	}
 
 
  /**
   * Add transitionInput to back of the Transition ArrayList
   * All observers are notified of this change (Model-View Architecture)
   *
   * @param transitionInput Transition Object to add
   */
  private void addTransition(Transition transitionInput) {
    boolean unique = true;
    if(transitionInput != null) {
      if(transitionInput.getId() != null && transitionInput.getId().length() > 0) {
        for(int i = 0 ; i < transitionsArray.size() ; i++) {
          if(transitionInput.getId().equals(((Transition)transitionsArray.get(i)).getId())) {
            unique = false;
          }
        }
      } else {
        String id = null;
        if(transitionsArray != null && transitionsArray.size() > 0) {
          int no = transitionsArray.size();
          do {
            // System.out.println("transition while loop");
            for(int i = 0 ; i < transitionsArray.size() ; i++) {
              id = "T" + no;
              if(transitionsArray.get(i) != null) {
                if(id.equals(((Transition)transitionsArray.get(i)).getId())) {
                  unique = false;
                  no++;
                } else {
                  unique = true;
                }
              }
            }
          } while(!unique);
        } else {
          id = "T0";
        }

        if(id != null) {
          transitionInput.setId(id);
        } else {
          transitionInput.setId("error");
        }
      }
      transitionsArray.add(transitionInput);
      setChanged();
      setMatrixChanged();
      //    notifyObservers(transitionInput.getBounds());
      notifyObservers(transitionInput);
    }
  }

  /**
   * Add arcInput to back of the Arc ArrayList
   * All observers are notified of this change (Model-View Architecture)
   *
   * @param arcInput Arc Object to add
   */
  public void addArc(Arc arcInput) {
    boolean unique = true;
    if(arcInput != null) {
      if(arcInput.getId() != null && arcInput.getId().length() > 0) {
        for(int i = 0 ; i < arcsArray.size() ; i++) {
          if(arcInput.getId().equals(((Arc)arcsArray.get(i)).getId())) {
            unique = false;
          }
        }
      } else {
        String id = null;
        if(arcsArray != null && arcsArray.size() > 0) {
          int no = arcsArray.size();
          do {
            for(int i = 0 ; i < arcsArray.size() ; i++) {
              id = "A" + no;
              if(arcsArray.get(i) != null) {
                if(id.equals(((Arc)arcsArray.get(i)).getId())) {
                  unique = false;
                  no++;
                } else {
                  unique = true;
                }
              }
            }
          } while(!unique);
        } else {
          id = "A0";
        }

        if(id != null) {
          arcInput.setId(id);
        } else {
          arcInput.setId("error");
        }
      }
      arcsArray.add(arcInput);
      addArcToArcsMap(arcInput);

      setChanged();
      setMatrixChanged();
      //notifyObservers(arcInput.getBounds());
      notifyObservers(arcInput);
    }
  }


  /** Update the arcsMap hashtable to reflect the new arc
   * @param arcInput New Arc
   * */
  private void addArcToArcsMap(Arc arcInput) {
    // now we want to add the arc to the list of arcs for it's source and target
    PlaceTransitionObject source = arcInput.getSource();
    PlaceTransitionObject target = arcInput.getTarget();
    ArrayList newList = null;


    if (source != null) {
// Pete: Place/Transitions now always moveable
//      source.setMovable(false);
      if (arcsMap.get(source)!=null) {
//        System.out.println("adding arc to existing list");
        ((ArrayList)arcsMap.get(source)).add(arcInput);
      } else {
//        System.out.println("creating new arc list");
        newList = new ArrayList();
        newList.add(arcInput);
        arcsMap.put(source,newList);
      }
    }

    if (target != null) {
// Pete: Place/Transitions now always moveable
//      target.setMovable(false);
      if (arcsMap.get(target)!=null) {
//        System.out.println("adding arc to existing list2");
        ((ArrayList)arcsMap.get(target)).add(arcInput);
      } else {
//        System.out.println("creating new arc list2");
        newList = new ArrayList();
        newList.add(arcInput);
        arcsMap.put(target,newList);
      }
    }

//    System.out.println("arcsMap size: " + arcsMap.size());

  }

  /* TK - this method is never called
   * Add tokenInput to the back of the Token ArrayList
   * All observers are notified of this change.
   *
   * @param tokenInput Token Object to add
   *
  private void addToken(Token tokenInput) {
    tokensArray.add(tokenInput);
    setChanged();
    setMatrixChanged();
    // notifyObservers(tokenInput.getBounds());
    notifyObservers(tokenInput);
  }
  */

  /**
   * Add any PetriNetObject - the object will be added to the appropriate list.
   * If the object passed in isn't a Transition, Place or Arc nothing will happen.
   * All observers are notified of this change.
   * @param pnObject The PetriNetObject to be added.
   */

  public void addPetriNetObject(PetriNetObject pnObject) {

    if (setPetriNetObjectArrayList(pnObject)) {


      if (pnObject instanceof Arc) {
        addArcToArcsMap((Arc)pnObject);
        addArc((Arc)pnObject);
      } else if (pnObject instanceof Place) {
        addPlace((Place)pnObject);
      } else if (pnObject instanceof Transition) {
        addTransition((Transition)pnObject);
      } else if (pnObject instanceof AnnotationNote) {
      	labelsArray.add(pnObject);
      }	else { // arrows, other labels.
        changeArrayList.add(pnObject);
        setChanged();
        setMatrixChanged();
        notifyObservers(pnObject);
      }

    }
    // we reset to null so that the wrong ArrayList can't get added to
    changeArrayList = null;
  }
//######################################################################################
  /**
   * Removes the specified object from the appropriate ArrayList of objects.
   * All observers are notified of this change.
   * @param pnObject The PetriNetObject to be removed.
   */

  public void removePetriNetObject(PetriNetObject pnObject){
    boolean didSomething = false;
    ArrayList attachedArcs = null;
    Iterator i = null;

    //System.out.println("removing: " + pnObject.getClass().getName());

    if (setPetriNetObjectArrayList(pnObject)) {
      didSomething = changeArrayList.remove(pnObject);


      // we want to remove all attached arcs also
      if (pnObject instanceof PlaceTransitionObject) {

        if ( (ArrayList)arcsMap.get(pnObject) != null) {

          // get the list of attached arcs for the object we are removing
          attachedArcs = ((ArrayList)arcsMap.get(pnObject));

          // remove the object we are removing from the arcsMap
          Arc removeArc;
          PlaceTransitionObject attached;

          // iterate over all the attached arcs, removing them all
          while(!attachedArcs.isEmpty()) {
            removeArc = (Arc)attachedArcs.get(0);
//            removePetriNetObject(removeArc);
            removeArc.delete();

          }
          arcsMap.remove(pnObject);
        }
      } else if (pnObject instanceof Arc) {

        // get source and target of the arc
        PlaceTransitionObject attached = ((Arc)pnObject).getSource();

          if (attached != null) {
            ArrayList a=(ArrayList)arcsMap.get(attached);
            if(a!=null)a.remove(pnObject);
          	attached.removeFromArc((Arc)pnObject);
          	if (attached instanceof Transition)
          		((Transition)attached).removeArcCompareObject((Arc)pnObject);
          	//attached.updateConnected(); ? causing null pointer exceptions
          }

          attached = ((Arc)pnObject).getTarget();

          if (attached != null) {
          	((ArrayList)arcsMap.get(attached)).remove(pnObject);
          	attached.removeToArc((Arc)pnObject);
          	if (attached instanceof Transition)
          		((Transition)attached).removeArcCompareObject((Arc)pnObject);
          	attached.updateConnected();
          }
      }      
      
      if (didSomething) {
        setChanged();
        setMatrixChanged();
        // notifyObservers(pnObject.getBounds());
        notifyObservers(pnObject);
      }
    }
    // we reset to null so that the wrong ArrayList can't get added to
    changeArrayList = null;
  }
    
  /* Returns an iterator for the transitions array.  Used by Animator.class to set all
       enabled transitions to highlighted */
  public Iterator returnTransitions(){
//    System.out.println("returnTransitionsIterator called");
    return transitionsArray.iterator();
  }

  /**
   * Sets an internal ArrayList according to the class of the object passed in.
   * @param pnObject The pnObject in question.
   * @return Returns True if the pnObject is of type Place, Transition or Arc
   */
  private boolean setPetriNetObjectArrayList(PetriNetObject pnObject) {

    // determine appropriate ArrayList
    if (pnObject instanceof Transition) {
      changeArrayList = transitionsArray;
      return true;
    } else if (pnObject instanceof Place){
      changeArrayList = placesArray;
      return true;
    } else if (pnObject instanceof Arc) {
      changeArrayList = arcsArray;
      return true;
    } else if(pnObject instanceof Token){
      changeArrayList = tokensArray;
      return true;
    } else if(pnObject instanceof AnnotationNote) {
      changeArrayList = labelsArray;
      return true;

    }
    return false;
  }

   /**
    * Returns an iterator of all PetriNetObjects - the order of these cannot be guaranteed.
    * @return An iterator of all PetriNetObjects
    */
  public Iterator getPetriNetObjects(){
    ArrayList all = new ArrayList(placesArray);
    all.addAll(transitionsArray);
    all.addAll(arcsArray);
    all.addAll(tokensArray);
    all.addAll (labelsArray);
    all.addAll(arrowsArray);

    return all.iterator();
  }

  /* TK - this method is never called
   * Set the Foward Incidence Matrix
   *
   * @param fowardIncidenceInput Foward Incidence Matrix object
   
  public void setFowardIncidenceMatrix(int[][] fowardIncidenceInput) {
    fowardsIncidenceMatrix = new PNMatrix(fowardIncidenceInput);
    fowardsIncidenceMatrix.matrixChanged = true;
  }
  */

  /* TK - this method is never called
   * Set the Backward Incidence Matrix
   *
   * @param backwardsIncidenceInput Backward Incidence Matrix object
   *

  public void setBackwardsIncidenceMatrix(int[][] backwardsIncidenceInput) {
    backwardsIncidenceMatrix = new PNMatrix(backwardsIncidenceInput);
    backwardsIncidenceMatrix.matrixChanged = true;
  }
  */

  /* TK - this method is never called
   * Set the Incidence Matrix
   *
   * @param incidenceInput Backward Incidence Matrix object
   *
  public void setIncidenceMatrix(int[][] incidenceInput) {
    incidenceMatrix = new PNMatrix(incidenceInput);
    incidenceMatrix.matrixChanged = true;
  }
  */

  /* TK - this method is never called
   * Set the Initial Markup Matrix
   *
   * @param markupInput Initial Markup Matrix object
   *
  public void setInitalMarkupMatrix(int[][] markupInput) {
    int i = 0;
    int j = 0;
    Integer intergerInput = null;
    for( ; i < markupInput.length ; i++) {
      fowardsIncidenceMatrix.add(i, new ArrayList(markupInput[i].length));
      for( ; j < markupInput[i].length ; j++) {
        intergerInput = new Integer(markupInput[i][j]);
        fowardsIncidenceMatrix.add(j, intergerInput);
      }
    }
  }
  */
  
  /**
   * Creates a Place object from a Place DOM Element
   *
   * @param inputPlaceElement Input Place DOM Element
   * @return Place Object
   */
  private Place createPlace(Element inputPlaceElement){
    double positionXInput = 0;
    double positionYInput = 0;
    String idInput = null;
    String nameInput = null;
    double nameOffsetYInput = 0;
    double nameOffsetXInput = 0;
    int initialMarkingInput = 0;
    double markingOffsetXInput = 0;
    double markingOffsetYInput = 0;

    String positionXTempStorage = inputPlaceElement.getAttribute("positionX");
    String positionYTempStorage = inputPlaceElement.getAttribute("positionY");
    String idTempStorage = inputPlaceElement.getAttribute("id");
    String nameTempStorage = inputPlaceElement.getAttribute("name");
    String nameOffsetYTempStorage = inputPlaceElement.getAttribute("nameOffsetX");
    String nameOffsetXTempStorage = inputPlaceElement.getAttribute("nameOffsetY");
    String initialMarkingTempStorage = inputPlaceElement.getAttribute("initialMarking");
    String markingOffsetXTempStorage = inputPlaceElement.getAttribute("markingOffsetX");
    String markingOffsetYTempStorage = inputPlaceElement.getAttribute("markingOffsetY");

    if (positionXTempStorage.length() > 0)
      positionXInput = Double.valueOf(positionXTempStorage).doubleValue()*(false ? DISPLAY_SCALE_FACTORX : 1)+(false ? DISPLAY_SHIFT_FACTORX : 1);
    if (positionYTempStorage.length() > 0)
      positionYInput = Double.valueOf(positionYTempStorage).doubleValue()*(false ? DISPLAY_SCALE_FACTORY : 1)+(false ? DISPLAY_SHIFT_FACTORY : 1);

    if (idTempStorage.length() > 0)
      idInput = idTempStorage;
    else if (nameTempStorage.length() > 0)
      idInput = nameTempStorage;

    if (nameTempStorage.length() > 0)
      nameInput = nameTempStorage;
    else if (idTempStorage.length() > 0)
      nameInput = idTempStorage;

    if (nameOffsetYTempStorage.length() > 0)
      nameOffsetXInput = Double.valueOf(nameOffsetYTempStorage).doubleValue();
    if (nameOffsetXTempStorage.length() > 0)
      nameOffsetYInput = Double.valueOf(nameOffsetXTempStorage).doubleValue();
    if (initialMarkingTempStorage.length() > 0)
      initialMarkingInput = Integer.valueOf(initialMarkingTempStorage).intValue();
    if (markingOffsetXTempStorage.length() > 0)
      markingOffsetXInput = Double.valueOf(markingOffsetXTempStorage).doubleValue();
    if (markingOffsetYTempStorage.length() > 0)
      markingOffsetYInput = Double.valueOf(markingOffsetYTempStorage).doubleValue();
    // Pete: Removed colour properties

    positionXInput+=PLACE_TRANSITION_HEIGHT/2-1;
    positionYInput+=PLACE_TRANSITION_HEIGHT/2-1;
    
    setMatrixChanged();
    return new Place(positionXInput, positionYInput, idInput, nameInput, nameOffsetXInput, nameOffsetYInput, initialMarkingInput, markingOffsetXInput, markingOffsetYInput);
  }

  /**
   * Creates a Label object from a Label DOM Element
   *
   * @param inputLabelElement Input Label DOM Element
   * @return Label Object
   */
  private AnnotationNote createAnnotation(Element inputLabelElement){
  	int positionXInput = 0;
  	int positionYInput = 0;
  	int widthInput = 0;
  	int heightInput = 0;
  	String text = null;
  	boolean borderInput = true;

  	String positionXTempStorage = inputLabelElement.getAttribute("xPosition");
  	String positionYTempStorage = inputLabelElement.getAttribute("yPosition");
  	String widthTemp = inputLabelElement.getAttribute("w");
  	String heightTemp = inputLabelElement.getAttribute("h");
  	String textTempStorage = inputLabelElement.getAttribute("txt");
  	String borderTemp = inputLabelElement.getAttribute("border");
  	
  	if (positionXTempStorage.length() > 0)
  		positionXInput = Integer.valueOf(positionXTempStorage).intValue()*(false ? DISPLAY_SCALE_FACTORX : 1)+(false ? DISPLAY_SHIFT_FACTORX : 1);
  	if (positionYTempStorage.length() > 0)
  		positionYInput = Integer.valueOf(positionYTempStorage).intValue()*(false ? DISPLAY_SCALE_FACTORX : 1)+(false ? DISPLAY_SHIFT_FACTORX : 1);
  	if (widthTemp.length() > 0)
  		widthInput = Integer.valueOf(widthTemp).intValue()*(false ? DISPLAY_SCALE_FACTORY : 1)+(false ? DISPLAY_SHIFT_FACTORY : 1);
  	if (heightTemp.length() > 0)
  		heightInput = Integer.valueOf(heightTemp).intValue()*(false ? DISPLAY_SCALE_FACTORY : 1)+(false ? DISPLAY_SHIFT_FACTORY : 1);
  	if (borderTemp.length()>0)
  		borderInput = Boolean.valueOf(borderTemp).booleanValue();
  	else borderInput = true;
  	if (textTempStorage.length() > 0)
  		text = textTempStorage;
  	else text = "";
  	
  	return new AnnotationNote(text, (int)positionXInput, (int)positionYInput, (int)widthInput, (int)heightInput, borderInput);
  }
  
  
  /**
   * Creates a Transition object from a Transition DOM Element
   *
   * @param inputTransitionElement Input Transition DOM Element
   * @return Transition Object
   */
  private Transition createTransition(Element inputTransitionElement){
    double positionXInput = 0;
    double positionYInput = 0;
    String idInput = null;
    String nameInput = null;
    double nameOffsetYInput = 0;
    double nameOffsetXInput = 0;
    double rate = 1.0;
    boolean timedTransition;
    int angle = 0;
    String positionXTempStorage = inputTransitionElement.getAttribute("positionX");
    String positionYTempStorage = inputTransitionElement.getAttribute("positionY");
    String idTempStorage = inputTransitionElement.getAttribute("id");
    String nameTempStorage = inputTransitionElement.getAttribute("name");
    String nameOffsetYTempStorage = inputTransitionElement.getAttribute("nameOffsetX");
    String nameOffsetXTempStorage = inputTransitionElement.getAttribute("nameOffsetY");
    String nameRate = inputTransitionElement.getAttribute("rate");
    String nameTimed = inputTransitionElement.getAttribute("timed");
    String nameAngle = inputTransitionElement.getAttribute("angle");
    if (nameTimed.length() == 0)
    	timedTransition = false;
    else if (nameTimed.length()== 5)
    	timedTransition = false;
    else
    	timedTransition = true;
    if (positionXTempStorage.length() > 0)
      positionXInput = Double.valueOf(positionXTempStorage).doubleValue()*(false ? DISPLAY_SCALE_FACTORX : 1)+(false ? DISPLAY_SHIFT_FACTORX : 1);
    if (positionYTempStorage.length() > 0)
      positionYInput = Double.valueOf(positionYTempStorage).doubleValue()*(false ? DISPLAY_SCALE_FACTORY : 1)+(false ? DISPLAY_SHIFT_FACTORY : 1);
    if (idTempStorage.length() > 0)
      idInput = idTempStorage;
    else if (nameTempStorage.length() > 0)
      idInput = nameTempStorage;
   
    if (nameTempStorage.length() > 0)
      nameInput = nameTempStorage;
    else if (idTempStorage.length() > 0)
      nameInput = idTempStorage;
   
    if (nameOffsetXTempStorage.length() > 0)
      nameOffsetXInput = Double.valueOf(nameOffsetXTempStorage).doubleValue();
   
    if (nameOffsetYTempStorage.length() > 0)
      nameOffsetYInput = Double.valueOf(nameOffsetYTempStorage).doubleValue();
   
    if (nameRate.length()== 0)
    	{	
    	 	nameRate = "1.0";
    	}
    if (nameRate != "1.0"){
    	rate = Double.valueOf(nameRate).doubleValue();}
    else 
    	rate = 1.0;

    if (nameAngle.length()>0)
    	angle = Integer.valueOf(nameAngle).intValue();
    // Pete: Removed colour properties
    positionXInput+=PLACE_TRANSITION_HEIGHT/2-1;
    positionYInput+=PLACE_TRANSITION_HEIGHT/2-1;
    setMatrixChanged();
    return new Transition(positionXInput, positionYInput,  idInput,  nameInput,  nameOffsetXInput, nameOffsetYInput, rate , timedTransition, angle);
  }

  /**
   * Creates a Arc object from a Arc DOM Element
   *
   * @param inputArcElement Input Arc DOM Element
   * @return Arc Object
   */
  private Arc createArc(Element inputArcElement){
    String idInput = null;
    String sourceInput = null;
    String targetInput = null;
    int weightInput = 1;
    double inscriptionOffsetXInput = 0;
    double inscriptionOffsetYInput = 0;
    double startX = 0;
    double startY = 0;
    double endX = 0;
    double endY = 0;
    
    
    sourceInput = inputArcElement.getAttribute("source");
    targetInput = inputArcElement.getAttribute("target");
    String idTempStorage = inputArcElement.getAttribute("id");
    String sourceTempStorage = inputArcElement.getAttribute("source");
    String targetTempStorage = inputArcElement.getAttribute("target");
    String inscriptionTempStorage = inputArcElement.getAttribute("inscription");
//    String inscriptionOffsetXTempStorage = inputArcElement.getAttribute("inscriptionOffsetX");
//    String inscriptionOffsetYTempStorage = inputArcElement.getAttribute("inscriptionOffsetY");


    if (idTempStorage.length() > 0)
      idInput = idTempStorage;
    if (sourceTempStorage.length() > 0)
      sourceInput = sourceTempStorage;
    if (targetTempStorage.length() > 0)
      targetInput = targetTempStorage;
    if (inscriptionTempStorage.length() > 0)
      weightInput = Integer.valueOf((inputArcElement.getAttribute("inscription") != null ? inputArcElement.getAttribute("inscription") : "1")).intValue();
//    if (inscriptionOffsetXTempStorage.length() > 0)
//      inscriptionOffsetXInput = Double.valueOf(inputArcElement.getAttribute("inscriptionOffsetX")).doubleValue();
//    if (inscriptionOffsetYTempStorage.length() > 0)
//     inscriptionOffsetYInput = Double.valueOf(inputArcElement.getAttribute("inscriptionOffsetY")).doubleValue();


    if (sourceInput.length() > 0) {
      if (getPlaceTransitionObject(sourceInput) != null) {
//        System.out.println("PNMLDATA: sourceInput is not null");
        startX = getPlaceTransitionObject(sourceInput).getPositionX();
        startY = getPlaceTransitionObject(sourceInput).getPositionY();
        if (CreateGui.getApp()!=null) {
        	startX += getPlaceTransitionObject(sourceInput).centreOffsetLeft();
        	startY += getPlaceTransitionObject(sourceInput).centreOffsetTop();
        }
      }
    }
    if (targetInput.length() > 0) {
      if (getPlaceTransitionObject(targetInput) != null) {
//        System.out.println("PNMLDATA: targetInput is not null");
        endX = getPlaceTransitionObject(targetInput).getPositionX();
        endY = getPlaceTransitionObject(targetInput).getPositionY();
      }
    }
    
    PlaceTransitionObject sourceIn = getPlaceTransitionObject(sourceInput);
    PlaceTransitionObject targetIn = getPlaceTransitionObject(targetInput);
    
    int aStartx;
    int aStarty;
    int aEndx;
    int aEndy;
    aStartx = sourceIn.getX();
    aStarty = sourceIn.getY();
    aEndx = targetIn.getX();
    aEndy = targetIn.getY();
    // add the insets and offset
    if (CreateGui.getApp()!=null) {
    	aStartx += sourceIn.centreOffsetLeft();
    	aStarty += sourceIn.centreOffsetTop();
    	aEndx += targetIn.centreOffsetLeft();
    	aEndy += targetIn.centreOffsetTop();
    }
   
    double _startx = aStartx;
    double _starty = aStarty;
    double _endx = aEndx;
    double _endy = aEndy;
    //TODO
    Arc tempArc = new Arc(_startx, _starty, _endx, _endy, sourceIn, targetIn, weightInput, idInput, Color.BLACK);
 
    getPlaceTransitionObject(sourceInput).addConnectFrom(tempArc);
    getPlaceTransitionObject(targetInput).addConnectTo(tempArc);
    
//**********************************************************************************
//The following section attempts to load and display arcpath details****************
//    ArcPath tempArcPath = tempArc.getArcPath();
    
    
    float arcPointX = 0;
    float arcPointY = 0;
    boolean arcPointType = false;
    String arcTempX = null;
    String arcTempY = null;
    String arcTempType = null;
    Node node = null;
    NodeList nodelist = null;
    nodelist = inputArcElement.getChildNodes();
    if (nodelist.getLength()>0) {
    	tempArc.getArcPath().purgePathPoints();
	    for (int i = 1; i < nodelist.getLength()-1; i++) {
	    	node = nodelist.item(i);
	    	if(node instanceof Element) {
		    	Element element = (Element)node;
		    	if ("arcpath".equals(element.getNodeName())){
		    		arcTempX = element.getAttribute("x");
		    		arcTempY = element.getAttribute("y");
		    		arcTempType = element.getAttribute("arcPointType");
		    		arcPointX = Float.valueOf(arcTempX).floatValue();
		    		arcPointY = Float.valueOf(arcTempY).floatValue();
            arcPointX+=ARC_CONTROL_POINT_CONSTANT+1;
            arcPointY+=ARC_CONTROL_POINT_CONSTANT+1;
		    		arcPointType = Boolean.valueOf(arcTempType).booleanValue();
		    		tempArc.getArcPath().addPoint(arcPointX,arcPointY,arcPointType);
		    		
		    	}
		    }
    	}
    }

//Arc path creation ends here***************************************************************    
//******************************************************************************************
    setMatrixChanged();
    return tempArc;
  }

  /**
   * Creates all Petri-Net Matrixes from current Petri-Net
   *
   */
  private void createMatrixes(){
	  createFowardIncidenceMatrix();
	  createBackwardsIncidenceMatrix();
	  createIncidenceMatrix();
	  createInitialMarkupMatrix();
	  createCurrentMarkupMatrix();
  }

  /**
   * Creates Foward Incidence Matrix from current Petri-Net
   *
   */
  private void createFowardIncidenceMatrix(){//m

	  int placeNo = 0, transitionNo = 0;
	  int placeSize = placesArray.size();
	  int transitionSize = transitionsArray.size();

	  Arc arc = null;
	  Place place = null;
	  Transition transition = null;

	  fowardsIncidenceMatrix = new PNMatrix(placeSize, transitionSize);

	  fowardsIncidenceMatrix.setToZero();

	  for(int i = 0 ; i < arcsArray.size() ; i++) {
		  arc = (Arc)arcsArray.get(i);
		  if(arc != null) {
			  PetriNetObject pnObject = arc.getTarget();
			  if(pnObject != null){
				  if(pnObject instanceof Place) {
					  place = (Place)pnObject;
					  pnObject = arc.getSource();
					  if(pnObject != null){
						  if(pnObject instanceof Transition){
							  transition = (Transition)pnObject;
							  transitionNo = getListPosition(transition);
							  placeNo = getListPosition(place);
							  fowardsIncidenceMatrix.set(placeNo, transitionNo, arc.getWeight());
						  }
					  }
				  }
			  }
		  }
	  }
	  fowardsIncidenceMatrix.matrixChanged = false;
  }

  /**
   * Creates Backwards Incidence Matrix from current Petri-Net
   *
   */
  private void createBackwardsIncidenceMatrix(){//Matthew

	  int placeNo = 0, transitionNo = 0;
	  int placeSize = placesArray.size();
	  int transitionSize = transitionsArray.size();
	  Arc arc = null;
	  Place place = null;
	  Transition transition = null;
	  backwardsIncidenceMatrix = new PNMatrix(placeSize, transitionSize);
	  backwardsIncidenceMatrix.setToZero();
	  for (int i = 0; i < arcsArray.size(); i++) {
		  arc = (Arc) arcsArray.get(i);
		  if (arc != null) {
			  PetriNetObject pnObject = arc.getSource();
			  if (pnObject != null) {
				  if (pnObject instanceof Place) {
					  place = (Place) pnObject;
					  pnObject = arc.getTarget();
					  if (pnObject != null) {
						  if (pnObject instanceof Transition) {
							  transition = (Transition) pnObject;
							  transitionNo = getListPosition(transition);
							  placeNo = getListPosition(place);
							  backwardsIncidenceMatrix.set(placeNo,
									  transitionNo, arc.getWeight());
						  }
					  }
				  }
			  }
		  }
	  }
	  backwardsIncidenceMatrix.matrixChanged  = false;
  }


  /**
   * Creates Incidence Matrix from current Petri-Net
   *
   */
  private void createIncidenceMatrix(){

	  createFowardIncidenceMatrix();
	  createBackwardsIncidenceMatrix();
	  incidenceMatrix = new PNMatrix(fowardsIncidenceMatrix.getArrayCopy());
	  incidenceMatrix = incidenceMatrix.minus(backwardsIncidenceMatrix);
	  incidenceMatrix.matrixChanged  = false;

  }

  /**
   * Creates Initial Markup Matrix from current Petri-Net
   *
   */
  private void createInitialMarkupMatrix(){

	  int placeNo = 0;
	  int placeSize = placesArray.size();
	  initialMarkupMatrix = new int[placeSize];
	  for (; placeNo < placeSize; placeNo++)
		  initialMarkupMatrix[placeNo] = ((Place) placesArray.get(placeNo))
		  .getInitialMarking();
	  initialMarkupMatrixChanged = false;
  }

  /**
   * Creates Initial Markup Matrix from current Petri-Net
   *
   */
  private void createCurrentMarkupMatrix(){

	  int placeNo = 0;
	  int placeSize = placesArray.size();
	  currentMarkupMatrix = new int[placeSize];
	  for (; placeNo < placeSize; placeNo++)
		  currentMarkupMatrix[placeNo] = ((Place) placesArray.get(placeNo))
		  .getCurrentMarking();
	  currentMarkupMatrixChanged = false;
  }

  /* TK - this method is never called
   * Stores Initial Markup Matrix from current Petri-Net Markup
   *
   *
  public void storeInitialMarking(){
    if (initialMarkupMatrixChanged) 
		createMatrixes();
	int placeNo = 0;
    int placeSize = placesArray.size();
    initialMarkupMatrix = new int[placeSize];
    for( ; placeNo < placeSize ; placeNo++)
      initialMarkupMatrix[placeNo] = ((Place)placesArray.get(placeNo)).getCurrentMarking();
  }
  */

  /* TK - this method is never called
   * Restores Initial Markup Matrix to current Petri-Net Markup
   *
   *
  public void restoreInitialMarking(){
    if (initialMarkupMatrixChanged) 
		createMatrixes();
	int placeNo = 0;
    int placeSize = placesArray.size();
    for( ; placeNo < placeSize ; placeNo++) {
      Place place = ((Place)placesArray.get(placeNo));
      if(place != null) {
        place.setCurrentMarking(initialMarkupMatrix[placeNo]);
        setChanged();
        setMatrixChanged();
        notifyObservers(place);
      }
    }
  }
  */

  /**
   * Stores Current Markup
   *
   */
  public void storeState(){
    int placeNo = 0;
    int placeSize = placesArray.size();
    markupMatrixAnimationStorage = new int[placeSize];
    for( ; placeNo < placeSize ; placeNo++)
      markupMatrixAnimationStorage[placeNo] = ((Place)placesArray.get(placeNo)).getCurrentMarking();
  }

  /**
   * Restores To previous Stored Markup
   */
  public void restoreState(){
    if(markupMatrixAnimationStorage != null) {
      int placeNo = 0;
      int placeSize = placesArray.size();
      for( ; placeNo < placeSize ; placeNo++) {
        Place place = ((Place)placesArray.get(placeNo));
        if(place != null) {
          place.setCurrentMarking(markupMatrixAnimationStorage[placeNo]);
          setChanged();
          notifyObservers(place);
          setMatrixChanged();
        }
      }
    }
  }

  /**
   * Fire a specified transition, no affect if transtions not enabled
   * @param transition Reference of specifiec Transition
   */
  public void fireTransition(Transition transition) {
    if(transition != null) {
      setEnabledTransitions();
      if(transition.isEnabled() && placesArray != null){
        int transitionNo = transitionsArray.indexOf(transition);
        for(int placeNo = 0 ; placeNo < placesArray.size() ; placeNo++) {
          ((Place)placesArray.get(placeNo)).setCurrentMarking((currentMarkupMatrix[placeNo] + incidenceMatrix.get(placeNo, transitionNo)));
        }
      }
    }
    setMatrixChanged();
  }

  /* TK - this method is never called
   * Fire a specified transition, no affect if transtions not enabled
   * @param transitionNo Position of Transition in internal ArrayList
   *
  private void fireTransition(int transitionNo) {
    setEnabledTransitions();
    if(((Transition)transitionsArray.get(transitionNo)).isEnabled() && placesArray != null){
      for(int placeNo = 0 ; placeNo < placesArray.size() ; placeNo++) {
        ((Place)placesArray.get(placeNo)).setCurrentMarking((currentMarkupMatrix[placeNo] + incidenceMatrix.get(placeNo, transitionNo)));
      }
    }
    setMatrixChanged();
  }
  */

  /**
   * This method will fire a random transition, and gives precedence 
   * to immediate transitions before considering "timed" transitions. 
   * The "rate" property of the transition is used as a weighting 
   * factor so the probability of selecting a transition is the 
   * rate of that transition divided by the sum of the weights of the
   * other enabled transitions of its class. The "rate" property can 
   * now be used to give priority among several enabled, immediate
   * transitions, or when there are no enabled, immediate transitions
   * to give priority among several enabled, "timed" transitions. 
   * 
   * Note: in spite of the name "timed" there is no probabilistic rate
   * calculated -- just a weighting factor among similar transitions. 
   * 
   * Changed by David Patterson Jan 2, 2006
   * 
   * Changed by David Patterson Apr 24, 2007 to clean up problems 
   * caused by fractional rates, and to speed up processing when only
   * one transition of a kind is enabled.
   * 
   * Changed by David Patterson May 10, 2007 to properly handle fractional
   * weights for immeditate transitions.
   * 
   * THe same logic is also used for timed transitions until the exponential
   * distribution is added. When that happens, the code will only be used for
   * immediate transitions.
   */
  public Transition fireRandomTransition() 
  {
	    Transition result=null;
	    Transition t;
	    setEnabledTransitions();
//	    int transitionsSize = transitionsArray.size()*transitionsArray.size()*transitionsArray.size();
	    int transitionNo = 0;

	    double rate = 0.0d;
	    double sumOfImmedWeights = 0.0d;
	    double sumOfTimedWeights = 0.0d;
	    ArrayList timedTransitions = new ArrayList();	// ArrayList<Transition>
	    ArrayList immedTransitions = new ArrayList();	// ArrayList<Transition>

	    for(transitionNo = 0 ; transitionNo < transitionsArray.size() ; transitionNo++)
	    {
	    	t = (Transition) transitionsArray.get(  transitionNo  );
    		rate = t.getRate();
	    	if( t.isEnabled())
	    	{
	    		if ( t.getTimed() )		// is it a timed transition
	    		{
    				timedTransitions.add( t );
    				sumOfTimedWeights += rate;
	    		}
	    		else					// immediate transition
	    		{
    				immedTransitions.add( t  );
    				sumOfImmedWeights += rate;
	    		}
	    	}		// end of if isEnabled
	    }		// end of for transitionNo

	    // Now, if there are immediate transitions, pick one
	    // next block changed by David Patterson to fix bug
	    int count = immedTransitions.size();
	    switch ( count )
	    {
	    	case 0:		// no immediate transitions
	    		break;	// skip out
	    	case 1: 	// only one immediate transition
	    		result = (Transition) immedTransitions.get( 0 );
	    		break;	// skip out
	    	default:	// several immediate transitions
	    		double rval = sumOfImmedWeights * randomNumber.nextDouble();
	    		for ( int index = 0; index < count; index++ )
	    		{
	    			t = (Transition) immedTransitions.get( index );
	    			rval -= t.getRate();
	    			if ( rval <= 0.0d )
	    			{
	    				result = t;
	    				break;
	    			}	
	    		}
	    }
	    if ( result == null )			// no immediate transition found
	    {
	        count = timedTransitions.size(); // count of timed, enabled transitions
	        switch( count )
	        {
	        	case 0:		// trouble! No enabled transition found
	        		break;
	        	case 1: 	// only one timed transition
	        		result = ( Transition ) timedTransitions.get( 0 );
	        		break;
	        	default:		// several timed transitions -- for now, pick one
	        		double rval = sumOfTimedWeights * randomNumber.nextDouble();
		    		for ( int index = 0; index < count; index++ )
		    		{
		    			t = (Transition) timedTransitions.get( index );
		    			rval -= t.getRate();
		    			if ( rval <= 0.0d )
		    			{
		    				result = t;
		    				break;
		    			}
		    		}
	        }
	    }

	    if ( result == null )
	    {
	        System.out.println( "no random transition to fire" );
	    }
	    resetEnabledTransitions();
	    return result;
  }     // end of method fireRandomTransition

  public void fireTransitionBackwards(Transition transition) {
    if(transition != null) {
      setEnabledTransitionsBackwards();
      if(transition.isEnabled() && placesArray != null){
        int transitionNo = transitionsArray.indexOf(transition);
        for(int placeNo = 0 ; placeNo < placesArray.size() ; placeNo++) {
          ((Place)placesArray.get(placeNo)).setCurrentMarking((currentMarkupMatrix[placeNo] - incidenceMatrix.get(placeNo, transitionNo)));
        }
      }
    }
    setMatrixChanged();
  }

  public void fireRandomTransitionBackwards() {
    setEnabledTransitionsBackwards();
    int transitionsSize = transitionsArray.size()*transitionsArray.size()*transitionsArray.size();
    int randomTransitionNumber = 0;
    Transition randomTransition = null;
    do {
      randomTransitionNumber = randomNumber.nextInt(transitionsArray.size());
      randomTransition = (Transition)transitionsArray.get(randomTransitionNumber);
      transitionsSize--;
      if(transitionsSize <= 0) break;
    } while(! randomTransition.isEnabled());
    fireTransitionBackwards(randomTransition);
//    System.out.println("Random Fired Transition Backwards" + ((Transition)transitionsArray.get(randonTransition)).getId());
  }

  /**
   * Determines whether all transitions are enabled and sets
   * the correct value of the enabled boolean
   */
  public void setEnabledTransitions() {

	  if (currentMarkupMatrixChanged) 
		  createMatrixes();
	  if(transitionsArray != null && placesArray != null) {
		  for(int transitionNo = 0 ; transitionNo < transitionsArray.size() ; transitionNo++) {
			  Transition transition = (Transition)transitionsArray.get(transitionNo);
			  if(transition != null) {
				  int test = placesArray.size();
				  for(int placeNo = 0 ; placeNo < placesArray.size() ; placeNo++) {
					  if(backwardsIncidenceMatrix.get(placeNo, transitionNo) <= currentMarkupMatrix[placeNo]) {
						  test--;
					  }
				  }
				  if(test <= 0) {
					  transition.setEnabled(true);
					  setChanged();
					  notifyObservers(transition);
				  } else {
					  transition.setEnabled(false);
					  setChanged();
					  notifyObservers(transition);
				  }
			  }
		  }
	  }
  }

  public void resetEnabledTransitions() {
	  if (currentMarkupMatrixChanged)   
		  createMatrixes();
	  if(transitionsArray != null && placesArray != null) {
		  for(int transitionNo = 0 ; transitionNo < transitionsArray.size() ; transitionNo++) {
			  Transition transition = (Transition)transitionsArray.get(transitionNo);
			  if(transition != null) {
				  int test = placesArray.size();
				  for(int placeNo = 0 ; placeNo < placesArray.size() ; placeNo++) {
					  if(backwardsIncidenceMatrix.get(placeNo, transitionNo) <= currentMarkupMatrix[placeNo]) {
						  test--;
					  }
				  }
				  if(test <= 0) {
					  transition.setEnabled(false);
				  } else {
					  transition.setEnabled(false);
				  }
				  setChanged();
				  notifyObservers(transition);
			  }
		  }
	  }
  }

  /**
   * Determines whether all transitions are enabled and sets
   * the correct value of the enabled boolean
   */
  private void setEnabledTransitionsBackwards() {
	  if (currentMarkupMatrixChanged) 
		  createMatrixes();
	  if(transitionsArray != null && placesArray != null) {
		  for(int transitionNo = 0 ; transitionNo < transitionsArray.size() ; transitionNo++) {
			  Transition transition = (Transition)transitionsArray.get(transitionNo);
			  if(transition != null) {
				  int test = placesArray.size();
				  for(int placeNo = 0 ; placeNo < placesArray.size() ; placeNo++) {
					  if(fowardsIncidenceMatrix.get(placeNo, transitionNo) <= currentMarkupMatrix[placeNo]) {
						  test--;
					  }
				  }
				  if(test <= 0) {
					  transition.setEnabled(true);
				  } else {
					  transition.setEnabled(false);
				  }
				  setChanged();
				  notifyObservers(transition);
			  }
		  }
	  }
  }

  /**
   * Empty all attributes, turn into empty Petri-Net
   */
  private void emptyPNML() {
    pnmlName = null;
    placesArray = null;
    transitionsArray = null;
    arcsArray = null;
    tokensArray = null;
    labelsArray = null;
    arrowsArray = null;
    changeArrayList = null;
    initialMarkupMatrix = null;
    markupHistoryMatrix = null;
    fowardsIncidenceMatrix = null;
    backwardsIncidenceMatrix = null;
    incidenceMatrix = null;
    arcsMap = null;
    initializeMatrixes();
  }

  /**
   * Get position of Petri-Net Object in ArrayList of given Petri-Net Object's type
   *
   * @param pnObject PlaceTransitionObject to get the position of
   * @return Position (-1 if not present) of Petri-Net Object in ArrayList of given Petri-Net Object's type
   */
  private int getListPosition(PetriNetObject pnObject){
    if (setPetriNetObjectArrayList(pnObject))
      return changeArrayList.indexOf(pnObject);
    else
      return -1;
  }

  /**
   * Get an List of all the Place objects in the Petri-Net
   *
   * @return An List of all the Place objects
   */
  public Place[] getPlaces() {
    Place[] returnArray = new Place[placesArray.size()];

    for(int i = 0 ; i < placesArray.size() ; i++)
      returnArray[i] = (Place)placesArray.get(i);

    return returnArray;
  }
  
  public int getPlacesCount(){
	  
	  if (placesArray == null) {
		  return 0;
	  } else {
		 
		  return placesArray.size();
	  
	  }	  
  }


  /**
   * Get a List of all the net-level NameLabel objects in the Petri-Net
   *
   * @return A List of all the net-level (as opposed to element-specific) label objects
   */
  public AnnotationNote[] getLabels() {
  	AnnotationNote[] returnArray = new AnnotationNote[labelsArray.size()];

  	for(int i = 0 ; i < labelsArray.size() ; i++)
  		returnArray[i] = (AnnotationNote)labelsArray.get(i);

  	return returnArray;
  }
  /**
   * Get an List of all the Transition objects in the Petri-Net
   *
   * @return An List of all the Transition objects
   */
  public Transition[] getTransitions() {
//    setEnabledTransitions();
    Transition[] returnArray = new Transition[transitionsArray.size()];

    for(int i = 0 ; i < transitionsArray.size() ; i++)
      returnArray[i] = (Transition)transitionsArray.get(i);

    return returnArray;
  }
  
  public int getTransitionsCount(){
	  
	  if (transitionsArray == null) {
		  return 0;
	  } else {
		 
		  return transitionsArray.size();
	  
	  }	  
  }

  /**
   * Get an List of all the Arcs objects in the Petri-Net
   *
   * @return An List of all the Arc objects
   */
  public Arc[] getArcs() {
    Arc[] returnArray = new Arc[arcsArray.size()];

    for(int i = 0 ; i < arcsArray.size() ; i++)
      returnArray[i] = (Arc)arcsArray.get(i);

    return returnArray;
  }

  /**
   * Return the Transition called transitionName from the Petri-Net
   *
   * @param transitionName Name of Transition object to return
   * @return The first Transition object found with a name equal to transitionName
   */
  public Transition getTransition(String transitionName) {
    Transition returnTransition = null;

    if(transitionsArray != null) {
      if(transitionName != null) {
        for(int i = 0 ; i < transitionsArray.size(); i++) {
          if(transitionName.equalsIgnoreCase(((Transition)transitionsArray.get(i)).getId())) {
            returnTransition = (Transition)transitionsArray.get(i);
          }
        }
      }
    }

    return returnTransition;
  }


  /**
   * Return the Transition called transitionName from the Petri-Net
   *
   * @param transitionNo No of Transition object to return
   * @return The Transition object
   */
  public Transition getTransition(int transitionNo) {
    Transition returnTransition = null;

    if(transitionsArray != null) {
      if(transitionNo < transitionsArray.size()) {
        returnTransition = (Transition)transitionsArray.get(transitionNo);
      }
    }

    return returnTransition;
  }

  /**
   * Return the Place called placeName from the Petri-Net
   *
   * @param placeName Name of Place object to return
   * @return The first Place object found with a name equal to placeName
   */
  public Place getPlace(String placeName) {
    Place returnPlace = null;

    if(placesArray != null) {
      if(placeName != null) {
        for(int i = 0 ; i < placesArray.size(); i++) {
          if(placeName.equalsIgnoreCase(((Place)placesArray.get(i)).getId())) {
            returnPlace = (Place)placesArray.get(i);
          }
        }
      }
    }

    return returnPlace;
  }

  /**
   * Return the Place called placeName from the Petri-Net
   *
   * @param placeNo No of Place object to return
   * @return The Place object
   */
  public Place getPlace(int placeNo) {
    Place returnPlace = null;

    if(placesArray != null) {
      if(placeNo < placesArray.size()) {
        returnPlace = (Place)placesArray.get(placeNo);
      }
    }

    return returnPlace;
  }

  /**
   * Return the Arc called arcName from the Petri-Net
   *
   * @param arcName Name of Arc object to return
   * @return The first Arc object found with a name equal to arcName
   */
  public Arc getArc(String arcName) {
    Arc returnArc = null;


    if(arcsArray != null) {
      if(arcsArray != null) {
        for(int i = 0 ; i < arcsArray.size(); i++) {
          if(arcName.equalsIgnoreCase(((Arc)arcsArray.get(i)).getId())) {
            returnArc = (Arc)arcsArray.get(i);
          }
        }
      }
    }

    return returnArc;
  }

  /* TK - this method is never called
   * Return the Arc called arcName from the Petri-Net
   *
   * @param arcName Name of Arc object to return
   * @return The first Arc object found with a name equal to arcName
   *
  public Arc getArcWithSource(PetriNetObject arcName) {
    Arc returnArc = null;

    if(arcsArray != null) {
      for(int i = 0 ; i < arcsArray.size(); i++) {
        if(arcsArray.get(i) != null) {
          Arc arc = (Arc)arcsArray.get(i);
          if(arc.getTarget() != null) {
            if(arcName.equals(arc.getSource())) {
              returnArc = arc;
            }
          }
        }
      }
    }

    return returnArc;
  }
  */

  /* TK - this method is never called
   * Return the Arc called arcName from the Petri-Net
   *
   * @param arcName Name of Arc object to return
   * @return The first Arc object found with a name equal to arcName
   *
  public Arc getArcWithTarget(PetriNetObject arcName) {
    Arc returnArc = null;

    if(arcsArray != null) {
      for(int i = 0 ; i < arcsArray.size(); i++) {
        if(arcsArray.get(i) != null) {
          Arc arc = (Arc)arcsArray.get(i);
          if(arc.getTarget() != null) {
            if(arcName.equals(arc.getTarget())) {
              returnArc = arc;
            }
          }
        }
      }
    }

    return returnArc;
  }
  */

  /**
   * Return the PlaceTransitionObject called ptoName from the Petri-Net
   *
   * @param ptoId Id of PlaceTransitionObject object to return
   * @return The first Arc PlaceTransitionObject found with a name equal to ptoName
   */
  public PlaceTransitionObject getPlaceTransitionObject(String ptoId) {

    if (ptoId != null) {
      if (getPlace(ptoId) != null)
        return getPlace(ptoId);
      else if (getTransition(ptoId) != null)
        return getTransition(ptoId);
    }

    return null;
  }

  /**
   * Return the Foward Incidence Matrix for the Petri-Net
   *
   * @return The Foward Incidence Matrix for the Petri-Net
   */
  public int[][] getForwardsIncidenceMatrix() {
	  if (fowardsIncidenceMatrix == null || fowardsIncidenceMatrix.matrixChanged) 
		  createFowardIncidenceMatrix();
	  return (fowardsIncidenceMatrix != null ? fowardsIncidenceMatrix.getArrayCopy() : null);
  }

  /**
   * Return the Backward Incidence Matrix for the Petri-Net
   *
   * @return The Backward Incidence Matrix for the Petri-Net
   */
  public int[][] getBackwardsIncidenceMatrix() {
	  if (backwardsIncidenceMatrix == null || backwardsIncidenceMatrix.matrixChanged) 
		  createBackwardsIncidenceMatrix();
	  return (backwardsIncidenceMatrix != null ? backwardsIncidenceMatrix.getArrayCopy() : null);
  }

  /**
   * Return the Incidence Matrix for the Petri-Net
   *
   * @return The Incidence Matrix for the Petri-Net
   */
  public int[][] getIncidenceMatrix() {
	  if (incidenceMatrix == null || incidenceMatrix.matrixChanged) 
		  createIncidenceMatrix();
	  return (incidenceMatrix != null ? incidenceMatrix.getArrayCopy() : null);
  }

  /**
   * Return the Initial Markup Matrix for the Petri-Net
   *
   * @return The Initial Markup Matrix for the Petri-Net
   */
  public int[] getInitialMarkupMatrix() {
	  if (initialMarkupMatrixChanged) 
		  createInitialMarkupMatrix();
	  return initialMarkupMatrix;
  }

  /**
   * Return the Initial Markup Matrix for the Petri-Net
   *
   * @return The Initial Markup Matrix for the Petri-Net
   */
  public int[] getCurrentMarkupMatrix() {
	  if (currentMarkupMatrixChanged) 
		  createCurrentMarkupMatrix();
	  return currentMarkupMatrix;
  }

  private void setMatrixChanged () {
	    
	    if (fowardsIncidenceMatrix != null) fowardsIncidenceMatrix.matrixChanged = true;
	    if (backwardsIncidenceMatrix != null) backwardsIncidenceMatrix.matrixChanged  = true;
	    if (incidenceMatrix != null) incidenceMatrix.matrixChanged = true;
	    initialMarkupMatrixChanged = true;
	    currentMarkupMatrixChanged = true;
	  
  }
  
  /**
	 * Create model from transformed PNML file
	 * 
	 * @author Ben Kirby, 10 Feb 2007
	 * @param filename URI location of PNML
	 * 
	 * @author Edwin Chung 
	 * This code is modified so that dataLayer objects can be created
	 * outside the GUI
	 */

	public void createFromPNML(Document PNMLDoc)
	{
		emptyPNML();
		Element element = null;
		Node node = null;
		NodeList nodeList = null;
		
		try {
			nodeList = PNMLDoc.getDocumentElement().getChildNodes();
			if (CreateGui.getApp()!=null) CreateGui.getApp().setMode(CREATING);				// Notifies used to indicate new instances.

			for(int i = 0 ; i < nodeList.getLength() ; i++) {
				node = nodeList.item(i);

				if(node instanceof Element) {
					element = (Element)node;
					if ("labels".equals(element.getNodeName())){
						addAnnotation(createAnnotation(element));
					}
					else if("place".equals(element.getNodeName())){ 
							addPlace(createPlace(element));
					}
					else if ("transition".equals(element.getNodeName())){
						addTransition(createTransition(element));
					}
					else if ("arc".equals(element.getNodeName())) {
						Arc a=createArc(element);
						addArc(a);
						a.addWeightLabelToContainer();
					}
				}
			}
			if (CreateGui.getApp()!=null) CreateGui.getApp().restoreMode();

		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 *
	 * Return a URI for the PNML file for the Petri-Net
	 *
	 * @return A DOM for the Petri-Net
	 */
	public String getURI() {
		return pnmlName;
	}


  /** prints out a brief representation of the dataLayer object */
  public void print() {
    System.out.println("No of Places = " + placesArray.size() + "\"");
    System.out.println("No of Transitions = " + transitionsArray.size() + "\"");
    System.out.println("No of Arcs = " + arcsArray.size() + "\"");
    System.out.println("No of Labels = " + labelsArray.size() + "\" (Model View Controller Design Pattern)");
  }
}


